/* miniDrafter version 0.07, Copyright (c) 2002 Krishna Jaidev S <mail@jaidev.port5.com>.

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
#include<dirent.h>
#include<unistd.h>
#include<stdlib.h>
#include<curses.h>
#include<termios.h>
#include<menu.h>
#include<string>
#include<list>
#include<fstream.h>
#include<panel.h>
#include<iostream.h>
#include<vector>
#include<sys/types.h>
#include<sys/stat.h>

#define BROWS_NEXT -4
#define LEFT -5
#define RIGHT -6
#define KEY_TAB 9

#define FILES 10
#define DIRS 11
#define INPUTFLD 12
#define OKBUT 13
#define CANBUT 14

#define EXIT_CODE 1000
#define KEY_ALT 27
#define KEY_SPACE 32
#define KEY_INS 331

#define OK -10
#define CANCEL -11

#define NEXT_MENU -3
#define PREV_MENU -4
#define NOT_SEL 100
#define END 10000

#define UNSAVED	-3
#define NEWFILE -4
#define SAVED 1
#define LAST -4

#define BG 0
#define FG 1

#define BKSP 1
#define DEL 2

#define NEXT 10
#define PREV -10

class window;
class buffer;
class openFile;

// global variables
int RESIZE = 0;
int TABSZ = 8;
int inpcolr;
int stdcolr;
int maincolr;
int menucolr;
int diacolr;
int butcolr;
int menusubcolr;
int INS = 1;					// flag for insert key 1 := insert, 0 := non insert
int PAIRNUM = 0;
string searchPattern;
string replacePattern;
string clipBoard;
WINDOW* mainWindow;
PANEL* mainPanel;
WINDOW* stdWin;
PANEL* stdPanel;

list <openFile *> fileList;
list <openFile *> :: iterator currentFile;
int numOpenFiles = 0;


int browse_driver (MENU * menu, WINDOW *sub, WINDOW* ipWin, vector <char *> flist);
string browse (string s, int & flg);

struct termios init_settings;	// save initial terminal settings and restore settings on exit
int initMainWin ();
void initProg ();
void terminateProg (int code);

int fileMenu (int *offItems, int numOffItems);
int editMenu (int *offItems, int numOffItems);
int searchMenu (int *offItems, int numOffItems);
int windMenu (int *offItems, int numOffItems);
int helpMenu (int *offItems, int numOffItems);
int displayMenu (ITEM ** items, int nItems, int offItems, int beginY, int beginX, int & ch);
int deleteMenu (int nItems, MENU * menu, ITEM ** items, WINDOW *fw, WINDOW *fsw, PANEL *fp, PANEL *fsp);
int noFilesMenu ();
int noFilesHandleCh (int ch);
int newFile();
int openFil ();
int goShell ();
int wait ();
int prompt (string &mesg);
int mesg (const char *msg);
string getIpStr (const char *mesg, int &flag, const char *def = 0);
int getIpInt (char * msg, int & flag);
int about();
int helpfile();
int cycleWindows (int dirn);
int saveAll ();

class window
{
	public:
		window (int nLines, int nCols, int bg = COLOR_BLUE, int fg = COLOR_WHITE);
		window () {}
		~window ();					
		int refreshPad ();
		int putChar (int ch, int y, int x, int nRowsBuf);
		int delCh(int y, int x, int ch, int key);
		int getChar ();
		int & cRow () { return _cRow; }
		int & cCol () { return _cCol; }
		WINDOW * pad () { return (_pad); }
		WINDOW * outerWin () { return (_outerWin); }
		int nRows () { return _nRows; }
		int nCols () { return _nCols; }
		int & cOrgX () {return _cOrgX; }
		int & cOrgY () {return _cOrgY; }
		int & oRows () { return _oRows;}
		int & oCols () {return _oCols; }
		int & pairnum () { return _pairnum;}
		 void updateXY ();
		 int resize();
		int refreshWin  ();
		int setXY (int y, int x);
		PANEL* outerPanel () {return _outerPanel;}
		int changeColor(int bgOrFg);
	private:
		WINDOW *_outerWin;
		int _bg;
		int _fg;
		int _fileNum;
		PANEL *_outerPanel;
		int _Y1, _Y2; // ymin and ymax of pad in terms of screen coordinates
		int _X1, _X2; // xmin and xmaz of pad in terms of screen coordinates
		int _beginX, _beginY;
		int _oRows, _oCols;
		WINDOW *_pad;
		int _nRows, _nCols;  // size of window
		int _cRow, _cCol;  // cursor position
		int _pairnum;
		int _cOrgX, _cOrgY;
};

class buffer
{
	public:
		buffer ();
		buffer (openFile & file, int & nRows);
		~buffer ();
		int insCh (int ch, int & nRows, int y = 0, int x = 0);
		int delCh (int y, int x, int & ch);
		int delLn (int y, int & nRows);
		int delWord (int &y, int &x);
		int insLn (int y);
		int saveBuffer(char *fileName, int nLines);
		int initBuffer(openFile & file, int & nRows);
		int numChars (int line);
		const char * cStr (int line);
		int mergeLines (int dest, int source);
		int lastChar (int line);
		int charCnt ();
		int wordCnt ();
		int find (string &pattern, int &y, int &x, int nRows, int beginY = -1, int beginX = -1);
		int replace (int y, int x, string & str, int n);
		int character (int i, int j);
	private:
		list <string> _fBuff;	// buffer (list of strings)
		list <string> :: iterator  _cRow;
		string :: iterator _cCol;			
		int _fileNum;
};

class openFile 
{
	public:
		openFile ();
		openFile (const char *fileName);
		~openFile ();					// Implement save unsaved file on exit here
		int insLn ();
		int insWord ();
		int delLn ();
		int delWord ();
		int readCh ();					// read from window
		int writeCh (int ch, int flag = 1);				// write to window and insert to buffer
		int handleCh (int ch);				// handle non display (spl) characters
		friend int handleAlt (openFile & fl);
		int deleteKey ();
		friend int work (openFile & fl);			 // Function that does most of the i/o
		int saveFile();
		int saveAs ();
		int closeFile();		
		friend int isdisplay (int ch);
		buffer & fileBuffer() { return _fileBuffer;}
		char* fileName() { return _fileName;}
		int backspKey ();
		int insertKey ();
		int navig (int ch);
		void updateXY();
		int nRows () {return _nRows;}
		window & fileWin() { return _fileWin;}
		int updatePad (int flg = 1);// updates content of pad call window :: refreshPad () after updatePad
		int charCnt() {return _fileBuffer.charCnt();}
		int wordCnt() {return _fileBuffer.wordCnt();}
		int find();
		int findAgain();
		int replace();
		int gotoLine();
		int & status () {return _status;}
		int beginMark ();
		int endMark ();
		int displayMark ();
		int eraseMark ();
		int cut ();
		int copy ();
		int paste ();
		int selectAll ();
		int delSelection ();
		int copyToClipBoard ();
		int MARKSTRT;
	private:
		char *_fileName;
		int _fileNum;
		class window _fileWin;		// window associated with file
						// row, col, INSERT, file name
		int _status;			// to be UNSAVED if file not yet saved;
		class buffer _fileBuffer;
		int _currentRow;
		int _currentCol;
		int _nRows;			// total number of lines in file;
		int markBegY, markBegX;
		int markEndY, markEndX;
		int markAnchorX, markAnchorY;

};

int openFile :: cut ()
{
	if (markBegX != -1) {
		clipBoard.erase (clipBoard.begin(), clipBoard.end());
		copyToClipBoard();
		delSelection();
		eraseMark();
	}
	return 1;
}

int openFile :: copy ()
{
	if (markBegX != -1) {
		clipBoard.erase (clipBoard.begin(), clipBoard.end());
		copyToClipBoard();
		eraseMark();
	}
	return 1;
}

int openFile :: paste ()
{
	int cr = _currentRow, cc = _currentCol;
	if (!clipBoard.length())
		return 0;
	int INSsave = INS;
	INS = 1;
	int i = 0;
	while (i < clipBoard.length() - 1)
		writeCh (clipBoard[i++], 0);
	INS = INSsave;
	_fileWin.cRow() = _currentRow = cr;
	_fileWin.cCol() = _currentCol = cc;
	updatePad(1);
	return 1;
}

int openFile :: selectAll ()
{
	markBegX = 0;
	markEndX = _fileBuffer.numChars(_nRows - 1) - 1;
	markBegY = 0;
	markEndY = _nRows - 1;
	displayMark();
	return 1;
}

int openFile :: copyToClipBoard ()
{
	
	int i = markBegY, ch, j = markBegX, k = markEndX;		
	if (markEndY == -1) return 0;
	while ((markBegY != markEndY && j < _fileBuffer.numChars(i) - 1) || (markBegY == markEndY && j <= k)) {
		clipBoard.push_back(_fileBuffer.character (i, j));
		j++;
	}
	if (clipBoard[clipBoard.length() - 1] != '\n')
		clipBoard.push_back('\n');
	i++;
	if (markBegY != markEndY) {
		while (i < markEndY) {
			clipBoard += string(_fileBuffer.cStr(i));
			if (clipBoard[clipBoard.length() - 1] != '\n')
				clipBoard.push_back('\n');
			i++;
		}
		j = 0;
		k = markEndX + 1;
		while (j <= k){
			clipBoard.push_back(_fileBuffer.character (i, j));
			j++;
		}
	}
	_fileWin.cRow() = _currentRow; //= markBegY;
	_fileWin.cCol() = _currentCol;// = markBegX;
	updatePad();
	return 1;
}
int openFile :: delSelection ()
{
	int i = markBegY, ch, k = markEndX;		
	if (markEndY == -1) return 0;
	while ((markBegY != markEndY && markBegX < _fileBuffer.numChars(i) - 1) || (markBegY == markEndY && markBegX < k)) {
		_fileBuffer.delCh (i, markBegX, ch);
		k --;
	}
	if (ch == '\n') _nRows --;
	i++;
	if (markBegY != markEndY) {
		while (i < markEndY) {
			_fileBuffer.delLn (i, _nRows);
			markEndY --;
		}
		int j = 0;
		k = markEndX;
		while (j <= k){
			_fileBuffer.delCh (i, j, ch);
			k --;
			if (ch == '\n')
				_nRows--;
		}
	}
	_fileWin.cRow() = _currentRow = markBegY;
	_fileWin.cCol() = _currentCol = markBegX;
	_fileBuffer.delCh (_currentRow, _currentCol, ch);
	if (ch == '\n')
		_nRows--;
	if (_nRows == 0) _nRows = 1;
	eraseMark();
	updatePad();
	return 1;
}
int openFile :: eraseMark ()
{
	markBegX = markEndX = markBegY = markEndY = markAnchorX = markAnchorY = -1;
	MARKSTRT = 0;
	return 1;
}

int openFile :: insWord ()
{

	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
	}
	if (_status == SAVED) _status = UNSAVED;
	update_panels();
	doupdate();
	updateXY();
	int flag;
	string temp = getIpStr("Enter Word: ", flag);
	if (flag == CANCEL)
		return 0;
	temp.push_back('\n');
	int i = 0, ch;
	while ((ch = temp[i++]) != '\n')
		writeCh(ch);
	writeCh (' ');
	return 1;
}

int openFile :: insLn ()
{
	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
	}
	if (_status == SAVED) _status = UNSAVED;
	_currentRow++;
	_currentCol = 0;
	if (_currentRow != _nRows)
		_fileBuffer.insCh ('\n', _nRows, _currentRow, _currentCol);
	else {
		_nRows++;
		_fileBuffer.insLn (_currentRow);
	}
	fileWin().setXY(1,0);
	fileWin().cCol() = 0;
	updatePad();
	return 1;	
}

int openFile :: delWord()
{
	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
	}
	int flag = _fileBuffer.delWord(_currentRow, _currentCol);
	_fileWin.cCol() = _currentCol;
	if (flag) {
		if (_status == SAVED) _status = UNSAVED;
		if (_fileBuffer.numChars(_currentRow) == 0)	
			delLn ();
		updatePad();
	}
	return 1;
}

int openFile :: delLn ()
{
	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
	}
	if (_status == SAVED) _status = UNSAVED;
	_fileBuffer.delLn(_currentRow, _nRows);
	updatePad();
	_currentCol = 0;
	fileWin().cCol() = 0;
	if (_currentRow == _nRows && _currentRow != 0) {
		fileWin().setXY(-1, 0);
		_currentRow--;
	}
/*	else if (_currentRow == 0 && _nRows == 0) {
		_fileBuffer.insCh ('\n', _nRows, 0, 0);
		_nRows = 1;
	}
*/
	updateXY();
	return 1;
	
}

int openFile :: saveAs ()
{
	updateXY();
	int flag;
	string name = browse ("Save As", flag);
//	string name = getIpStr ("Save As", flag);
	if (flag == CANCEL) 
		return 0;
	if ((flag = _fileBuffer.saveBuffer ((char*) name.c_str(), _nRows)) != -1) {
		mvwhline (fileWin().outerWin(), 0, 1, ACS_HLINE, fileWin().oCols() - 2);
		delete [] _fileName;
		_fileName = new char [strlen (name.c_str()) + 1];
		strcpy (_fileName, name.c_str());
		_status = SAVED;
	}
	updateXY();
}

int openFile :: saveFile()
{
	int retVal = 1;
	if (_status == NEWFILE)
		saveAs();
	else {
		int flag = fileBuffer().saveBuffer(fileName(), nRows());
		if (flag) _status = SAVED;
		else retVal = -1;
	}
	updateXY();
	return retVal;
}

int openFile :: updatePad (int flag)
{
	int i;
	for (i = 0; i < _nRows && _fileBuffer.cStr(i) != 0; i++) {
		if (flag == 2 && i == 40) {
			updateXY();
			flag = 1;
		}
		mvwprintw(_fileWin.pad(), i, 0, "%-*s", _fileWin.oCols(), _fileBuffer.cStr(i));
		wclrtoeol(_fileWin.pad());
	}
	wclrtobot(_fileWin.pad());
	if (markBegX != -1)
		displayMark();
	if (flag == 1)updateXY();
	return 1;		

}

int openFile :: beginMark () 
{
	markBegY = markAnchorY = _currentRow;
	markBegX = markAnchorX = _currentCol;
	MARKSTRT = 1;
	return 1;
}

int openFile :: endMark () 
{
	MARKSTRT = 0;
	return 1;
}

int openFile :: displayMark ()
{
	int i = markBegY, j = markBegX;
	wattron (_fileWin.pad(), A_STANDOUT);
	while ((markBegY == markEndY && j <= markEndX && j < _fileBuffer.numChars(i)) || (i != markEndY && j < _fileBuffer.numChars (i) - 1)) {
		mvwaddch(_fileWin.pad(), i, j, _fileBuffer.character (i, j));
		j++;
	}
	if (i != markEndY) {
		i++;
		for (; i != markEndY && i < markEndY && _fileBuffer.cStr(i) != 0; i++) {
			mvwprintw(_fileWin.pad(), i, 0, "%s",_fileBuffer.cStr(i));
		}
		j = 0;
		while (j <= markEndX && j < _fileBuffer.numChars(i)) {
			mvwaddch(_fileWin.pad(), i, j, _fileBuffer.character (i, j));
			j++;
		}
	}
	wattroff (_fileWin.pad(), A_STANDOUT);
	return 1;
}

int handleAlt (openFile & fl)
{
	int ch;
	ch = fl.readCh();
	int retVal = 0, i = 0;
	int off[3];
	sw:
	switch (ch) {
		case 'm':
		case 'M':
			if (fl.MARKSTRT) {
			//	fl.endMark();
				fl.eraseMark();
				fl.updatePad();
			}
			else
				fl.beginMark();
			break;	
		case 'c':	
		case 'C':
			fl.copy();
			break;
		case 'x':
		case 'X':
			fl.cut();
			break;
		case 'v':
		case 'V':
			fl.paste();
			break;
		case 'a':
		case 'A':
			fl.selectAll();
			break;
		case 'i':
		case 'I':
			fl.insWord();
			break;
		case 'G':
		case 'g':
			fl.gotoLine();
			break;
		case 'S':
		case 's' :
			off[0] = 3;
			retVal = searchMenu (off, 1);
			break;
		case 'f' :
		case 'F' :
			off[0] = 5;
			off[1] = 7;
			off[2] = 9;
			retVal = fileMenu(off, 3);
			break;
		case 'e' :
		case 'E' :
			off[0] = 5;
			off[1] = 10;
			retVal = editMenu(off, 2);
			break;
		case 'w' :
		case 'W' : 
			i = 0;
			if (numOpenFiles == 1) {
				off[i++] = 1;
				off[i++] = 2;
			}
			off[i++] = 3;
			retVal = windMenu(off, i);
			break;
		case 'h' :
		case 'H' :
			retVal = helpMenu(off, 0);
			break;
		case 't':
		case 'T':
			fl.delWord();
			break;
		case 'y':
		case 'Y':
			fl.delLn();
			break;
		case 'L':
		case 'l':
			fl.insLn();
			break;
		default:
			;
	}
	fl.updateXY();
	if (retVal == 6) retVal = 0;
	if (retVal != 0) {
		ch = retVal;
		goto sw;
	}
	return 1;
}

int openFile :: handleCh (int ch)
{
	int y = 0, x = 0;
	if (ch == -1) return -1;
	string mesgstr;
	switch (ch) {
		case KEY_ALT : 
			return (handleAlt(*this));
		case KEY_BACKSPACE :
			backspKey ();
			break;
		case KEY_DC:
			deleteKey ();
			break;
		case KEY_RESIZE:
			beep();
			break;
		case KEY_F(8):
			_fileWin.resize();
			updatePad();
			break;
		case KEY_F(9):
			cycleWindows(PREV);
			break;
		case KEY_F(10):
			cycleWindows(NEXT);
			break;
		case KEY_F(12):
			terminateProg(0);
		case KEY_F(2):
			saveFile();
			break;
		case KEY_F(1):
			newFile();
			break;
		case KEY_F(11):
			goShell();
			updateXY();
			break;
		case KEY_F(4):
			closeFile();
			break;
		case KEY_F(3):
			openFil();
			break;
		case KEY_PPAGE:
		case KEY_NPAGE:
		case KEY_END :
		case KEY_HOME :
		case KEY_RIGHT : 
		case KEY_LEFT :
		case KEY_UP :
		case KEY_DOWN :
			navig (ch);
			break;
		case KEY_INS:
			insertKey ();			
			break;		
		case KEY_F(5):
			find();
			break;
		case KEY_F(6):
			findAgain();
			break;
		case KEY_F(7):
			replace();
			break;
		default :
			mesgstr.push_back ('\n');
			mesgstr += string ("  Unknown command :(");
			mesgstr.push_back('\n');
			prompt(mesgstr);	
			updatePad();
	}
	return 1;
}

int openFile :: navig (int ch)
{
	int newRow = _currentRow;
	int newCol = _currentCol;
	int oldRow = _currentRow;
	int oldCol = _currentCol;
	int changerow = 0;
	int changecol = 0;
	switch (ch) {
		case KEY_UP:
			newRow -= 1;
			changerow = 1;
			break;
		case KEY_DOWN:
			newRow += 1;
			changerow = 1;
			break;
		case KEY_RIGHT:
			newCol += 1;
			changecol = 1;
			break;
		case KEY_LEFT:
			newCol -= 1;
			changecol = 1;
			break;
		case KEY_HOME:
			newCol = 0;
			changecol = 1;
			break;
		case KEY_END:
			newCol = _fileBuffer.numChars(_currentRow) - 1;
			changecol = 1;
			break;
		case KEY_NPAGE:
			newRow += _fileWin.oRows() - 2;
			changerow = 1;
			break;
		case KEY_PPAGE:
			newRow -= _fileWin.oRows() - 2;
			changerow = 1;
			break;
	}
	if (newRow >= _nRows)
		newRow = _nRows - 1;
	if (newRow < 0) newRow = 0;
	if (newRow < _nRows && newRow >= 0 &&  newCol >= 0) {
		if (newCol >= _fileBuffer.numChars(newRow) - 1) {
			newCol = _fileBuffer.numChars(newRow) - 1;
			if ((ch == KEY_END || ch == KEY_RIGHT || changerow) && _fileBuffer.lastChar (newRow) != '\n')
				newCol++;
		}
		_currentRow = newRow;
		_currentCol = newCol;
		updateXY();
	}
	else 
		beep();
	if (MARKSTRT) {
		if (_currentRow > markAnchorY || (_currentRow == markAnchorY && _currentCol > markAnchorX)){
			markEndY = _currentRow;
			markEndX = _currentCol;
			markBegX = markAnchorX;
			markBegY = markAnchorY;
		}
		else {
			markBegY = _currentRow;
			markBegX = _currentCol;
			markEndX = markAnchorX;
			markEndY = markAnchorY;
		}
		updatePad();
	}	
	return (1);	
}


int work (openFile & fl)
{
	fl.updateXY();
	int ch;
	int flag; 
	while (1) {
		if (numOpenFiles)
			ch = fl.readCh ();
		if (isdisplay (ch))
			fl.writeCh (ch);
		else if ((flag = fl.handleCh (ch)) == EXIT_CODE) 
			return EXIT_CODE;
	}
	return 1;
}


openFile :: openFile () : _fileWin (4 * (LINES - 4), 4 * (COLS - 2), COLOR_BLUE, COLOR_WHITE), _fileBuffer()
{
	_fileName = new char [11];
	_fileNum = numOpenFiles;
	strcpy (_fileName, "UNTITLED");
	_fileName[8] = '0' + numOpenFiles / 10;
	_fileName[9] = '0' + numOpenFiles % 10;
	_fileName[10] = '\0';
	_currentRow = _currentCol = 0;
	_status = NEWFILE;	
	MARKSTRT = 0;
	_nRows = 1;
	updateXY();
	markBegX = markEndX = markBegY = markEndY = markAnchorX = markAnchorY = -1;
}


openFile :: openFile (const char *fileName)  :_fileWin (4 * (LINES - 4), 4 * (COLS - 2),  COLOR_BLUE, COLOR_WHITE)
{	_fileName = new char [strlen (fileName) + 1];
	strcpy (_fileName, fileName);	
	_status = SAVED;	
	_fileNum = numOpenFiles;
	_currentRow = 0;
	_currentCol = 0;
	_fileBuffer = buffer(*this, _nRows);
	_fileWin.cRow() = 0;
	_fileWin.cCol() = 0;
	MARKSTRT = 0;
	markBegX = markEndX = markBegY = markEndY = markAnchorX = markAnchorY = -1;
	updatePad(2);
}

void openFile :: updateXY ()
{
	mvwprintw (_fileWin.outerWin(), 0 , _fileWin.oCols() / 2 - strlen (_fileName) / 2 - 3, "[%s]", _fileName);	
	if (RESIZE) {
		mvwprintw (stdWin, LINES - 1, 0, "            ");
		mvwaddch (stdWin, LINES -1, 1, ACS_UARROW);
		mvwaddch (stdWin, LINES -1, 3, ACS_DARROW);
		mvwaddch (stdWin, LINES -1, 5, ACS_LARROW);
		mvwaddch (stdWin, LINES -1, 7, ACS_RARROW);
		mvwprintw (stdWin, LINES - 1, 9, "%-*s", COLS - 19, " - RESIZE | F - Full Screen");
		mvwprintw (stdWin, LINES - 1,COLS - 9, "RESIZE ");
	}
	else if (MARKSTRT) {
		mvwprintw (stdWin, LINES - 1, 1, "%-*s", COLS - 12, "ALT M - Normal Mode, Alt X - Cut, Alt C - Copy, Alt V - Paste");
		mvwprintw (stdWin, LINES - 1,COLS - 9, "  MARK ");
	}
	else if (INS) {
		curs_set (1);
		mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT ");
		if (markBegX != -1)
			mvwprintw (stdWin, LINES - 1, 1, "%-*s", COLS - 12, "Alt X - Cut, Alt C - Copy, Alt V - Paste");
		else
			mvwprintw (stdWin, LINES - 1, 1,  "%-*s", COLS - 12,  "F1 - New | F2 - Save | F3 - Open | F11 - Shell | F12 - Exit");
	}
	else {
		curs_set (2);
		mvwprintw (stdWin, LINES - 1,COLS - 9, "REPLACE");
		if (markBegX != -1)
			mvwprintw (stdWin, LINES - 1, 1, "%-*s", COLS - 12, "Alt X - Cut, Alt C - Copy, Alt V - Paste");
		else
			mvwprintw (stdWin, LINES - 1, 1,  "%-*s", COLS - 12,  "F1 - New | F2 - Save | F3 - Open | F11 - Shell | F12 - Exit");
	}
	
	if (_status == UNSAVED || _status == NEWFILE) 
		waddch (_fileWin.outerWin(), '*');
	else
		whline (_fileWin.outerWin(), ACS_HLINE, 1);
	while ((_currentCol - _fileWin.cOrgX()) >= _fileWin.oCols() - 2)
		_fileWin.cOrgX() += 1;
	while (_currentCol < _fileWin.cOrgX())
		_fileWin.cOrgX() -= 1;
	while (_currentRow - _fileWin.cOrgY() >= _fileWin.oRows() - 2)
		_fileWin.cOrgY() += 1;
	while (_currentRow < _fileWin.cOrgY())
		_fileWin.cOrgY() -= 1;
	if (_fileWin.cOrgY () > _nRows) 
		_fileWin.cOrgX () = _nRows - _fileWin.oRows() + 2;
	int here;
	if (_fileWin.oRows() > 4) {
		here = float (_currentRow)/_nRows * (_fileWin.oRows () - 4) + 2;
		if (here > (_fileWin.oRows() - 4))
			 here = _fileWin.oRows() - 3;
		mvwaddch (_fileWin.outerWin(), 1, _fileWin.oCols() - 1, ACS_UARROW);
		mvwvline (_fileWin.outerWin(), 2, _fileWin.oCols() - 1, ACS_CKBOARD, _fileWin.oRows() - 3);
		mvwaddch (_fileWin.outerWin(), _fileWin.oRows() - 2, _fileWin.oCols() - 1, ACS_DARROW);
		mvwaddch (_fileWin.outerWin(), here, _fileWin.oCols() - 1, ACS_BLOCK);
	}
	_fileWin.cRow () = _currentRow;
	_fileWin.cCol () = _currentCol;
	_fileWin.updateXY();
}


int openFile :: insertKey ()
{
	INS = !INS;
	if (INS)
		mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT");
	else 
		mvwprintw (stdWin, LINES - 1, COLS - 9, "      ");
	updateXY();
	return 1;
}

openFile :: ~openFile ()
{
	;
}

int openFile :: deleteKey ()
{
	if (_status == SAVED) _status = UNSAVED;
	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
		return 1;
	}
	int flag;
	int ch;
	flag = _fileBuffer.delCh (_currentRow, _currentCol, ch);
	if (flag == 2 && ch == '\n'){	
		_nRows --;
		updatePad();
	}
	else if (flag == 3 && _nRows > 1) {
		//_nRows--;
	}
	else if (flag == 1) {
		_fileWin.delCh(_currentRow, _currentCol, ch, DEL);
	}
	else beep();
	updateXY();
	return 1;
}

int openFile :: backspKey ()
{
	if (_status == SAVED) _status = UNSAVED;
	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
		return 1;
	}
	_currentCol--;
	int ch;
	if (_currentCol != -1) {
		_fileBuffer.delCh (_currentRow, _currentCol, ch);
		_fileWin.delCh(_currentRow, _currentCol, ch, BKSP);
	}
	else if (_currentRow > 0) {
		_currentRow--;	
		_currentCol = _fileBuffer.numChars(_currentRow);
		if (_fileBuffer.mergeLines (_currentRow, _currentRow + 1))
			_currentCol--;
		_nRows--;
		_fileWin.setXY (-1, 0);
		_fileWin.cCol() = _currentCol;
		updatePad();
	}
	else {
		_currentCol++;
		beep();
	}
	updateXY();
	return 1;	
}
int openFile :: readCh ()
{
	if (INS) curs_set (1);
	else curs_set (2);
	int ch = _fileWin.getChar();
	return (ch);
}

int openFile :: writeCh (int ch, int flag)
{
	if (markBegX != -1) {
		delSelection();
		eraseMark();
		updatePad();
	}
	if (_status == SAVED) _status = UNSAVED;
	if (ch != '\t'){
		_fileBuffer.insCh (ch, _nRows, _currentRow, _currentCol);
		_fileWin.putChar (ch, _currentRow, _currentCol++, _nRows);
		if (ch == '\n') {
			_currentCol = 0;
			if (flag) updatePad(flag);
			_currentRow++;
		}
	}
	if (ch == '\t') {
		int nBlks;
		nBlks = TABSZ - (_currentCol % TABSZ);
		for (int i = 0; i < nBlks ; i++) {
			_fileBuffer.insCh (' ', _nRows, _currentRow, _currentCol);
			_fileWin.putChar(' ', _currentRow, _currentCol++, _nRows);
		}
	}			
	if (flag)
		updateXY();
	return 1;
}

int openFile :: closeFile ()
{
	if (numOpenFiles) {
		currentFile = fileList.begin();
		if ((_status == NEWFILE && _fileBuffer.charCnt() > 1)|| _status == UNSAVED) {
			string newstr;
			newstr.push_back('\n');
			newstr += string (" Warning : File not Saved. Save ?");
			int flag = prompt (newstr);
			newstr.erase (newstr.begin(), newstr.end());
			if (flag == OK) {
				if (_status == NEWFILE)
					saveAs();
				else 
					saveFile();
			}
		}
		delete (*currentFile);
		fileList.erase(currentFile);
		numOpenFiles--;
	}
	else return 0;
	if (numOpenFiles) {
		currentFile = fileList.begin();
		mvwprintw ((*currentFile)->fileWin().outerWin(), 0, (*currentFile)->fileWin().oCols() / 2 - strlen ((*currentFile)->fileName()) / 2 - 3, "[%s]", (*currentFile)->fileName());
		if (INS)
			mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT");
		else 
			mvwprintw (stdWin, LINES - 1, COLS - 9, "      ");
		(*currentFile)->updatePad();
		work (**currentFile);
	}
	else {
		mvwprintw (stdWin, LINES - 1, 1,  "%-*s", COLS - 12,  "F1 - New | F2 - Save | F3 - Open | F11 - Shell | F12 - Exit");
		update_panels();
		doupdate();
		wait();
	}
	return 1;
}


int openFile :: find() 
{	
	int flag;
	updateXY();
	string pattern = getIpStr ("Search pattern :", flag, searchPattern.c_str());
	searchPattern = pattern;
	if (flag == CANCEL)
		return 0;
	int x, y; 
	flag = _fileBuffer.find(pattern, y, x, _nRows);
	if (flag == 0) {
		mesg ("Search pattern not found :(");
		updateXY();
		return 0;
	}
	_currentRow = y;
	_currentCol = x;
	_fileWin.cRow() = _currentRow;
	_fileWin.cCol() = _currentCol;	
	markBegY = _currentRow;
	markEndY = _currentRow;
	markBegX = _currentCol;
	markEndX = _currentCol + pattern.length() - 1;
	updatePad();
	return 1;
};


int openFile :: findAgain() 
{
	updateXY();
	if (searchPattern.length() == 0) {
		beep();
		return 0;
	}
	int x, y, flag; 
	flag = _fileBuffer.find(searchPattern, y, x, _nRows, _currentRow, _currentCol);
	if (flag == 0) {
		mesg ("Search pattern not found :(");
		updateXY();
		return 0;
	}
	_currentRow = y;
	_currentCol = x;
	_fileWin.cRow() = _currentRow;
	_fileWin.cCol() = _currentCol;	
	markBegY = _currentRow;
	markEndY = _currentRow;
	markBegX = _currentCol;
	markEndX = _currentCol + searchPattern.length() - 1;
	updatePad();
	return 1;
}

int openFile :: replace() 
{
	int findflag = find(), flag;	
	if (!findflag) return 0;
	string repstr = getIpStr ("Replace pattern :", flag, replacePattern.c_str());
	replacePattern = repstr;
	updateXY();
	if (flag == CANCEL)
		return 0;
	flag = OK;
	string msg = ("Replace this occurance ?");
	here:
	do {
		flag = prompt (msg);
		if (flag == CANCEL) 
			findflag = findAgain();
	} while (flag == CANCEL && findflag != 0);
	if (findflag == 0)
		return 0;
	_fileBuffer.replace (_currentRow, _currentCol, repstr, searchPattern.length());
	markBegY = _currentRow;
	markEndY = _currentRow;
	markBegX = _currentCol;
	markEndX = _currentCol + replacePattern.length() - 1;
	updatePad();
	string msg2 = string ("Replace Again ?");
	flag = prompt (msg2);
	if (flag == OK && findflag != 0) {
		findflag = findAgain();
		if (findflag)
			goto here;	
	}		
	eraseMark();
	updateXY();
	
	return 1;
}

int openFile :: gotoLine() 
{
	updateXY();
	int flag = OK, linenum = getIpInt ("Line Number:", flag); 
	if (flag == CANCEL || linenum > _nRows || linenum <= 0) {
		beep(); 
		return 0;
	}
	_currentRow = linenum - 1;
	_currentCol = 0;
	_fileWin.cOrgX() = 0;
	_fileWin.cRow() = _currentRow;
	_fileWin.cCol() = 0;
	updateXY();
	return 1;
}

//-------------------------------------------------------------------------------------------


int buffer :: charCnt ()
{
	list <string> :: iterator temp;
	temp = _fBuff.begin();
	int cnt = 0;
	while (temp != _fBuff.end()) {
		cnt += temp->length();
		temp++;
	}
	return cnt;
}

int buffer :: wordCnt ()
{
	list <string> :: iterator temp;
	string :: iterator stemp;
	temp = _fBuff.begin();
	int cnt = 0;
	while (temp != _fBuff.end()) {
		if (temp->length()  == 1) cnt--;
		stemp = temp->begin();
		while (stemp != temp->end()){
			if ((*stemp) == ' '||(*stemp) == '\t' || (*stemp) == '\n') {
				cnt++;
				stemp++;
				while ((*stemp) == ' ' && stemp != temp->end())
					stemp++;
			}
			else
				stemp++;
		}
		temp++;
	}
	return cnt;
}


int buffer :: insLn (int y)
{
	_fBuff.push_back("");
	list <string> :: iterator temp = _fBuff.begin();
	for (int i = 0; i < y; i++, temp++);
	temp->push_back('\n');
	return 1;
}

int buffer :: delWord(int &y, int &x) 
{
	list <string> :: iterator temp = _fBuff.begin();
	int i = 0;
	while (i < y) {
		i++;
		temp++;
	}
	int begin, end;
	begin = end = x;
	string :: iterator stemp = temp->begin() + x;
	if (x != 0 && x != temp->length() - 1) {
		while ((*stemp) != ' ') {
			stemp --;
			begin --;
		}
	}
	else if (!x)
		begin = 0;
	else 
		begin = temp->length() - 1;
	stemp = temp->begin() + x;
	if (x < temp->length() - 1) {
		while ((*stemp) != ' ' && end < temp->length()) {
			stemp++;
			end++;
		}
	}
	else
		end = temp->length() - 1;
	if (begin <= 0)
		begin = 0;
	else
		begin++;
	if (end >= temp->length() - 1) 
		end = temp->length() - 1;
	if (begin < end)
		temp->erase (temp->begin() + begin, temp->begin() + end + 1);
	else {
		beep();
		return -1;
	}
	if (end == temp->length() - 1) 
		temp->push_back('\n');
	if (x >= temp->length() - 1)
		x = temp->length();
	else x = begin;
	return end - begin;
}

int buffer :: character (int y, int x)
{				// zero based indices
	list <string> :: iterator temp = _fBuff.begin();
	int i = 0;
	while (i < y) {
		temp++;
		i++;
	}
	return (*temp)[x];
	
}

const char* buffer :: cStr (int line) {
	list <string> :: iterator temp;
	temp = _fBuff.begin();
	for (int i = 0; i <= line && temp != _fBuff.end(); ++temp, i++);
	temp--;
	if (temp != _fBuff.end()) return temp->c_str();
	else return 0;
}

int buffer :: mergeLines (int dest, int source)
{
	int retval = 0;
	list <string> :: iterator temp = _fBuff.begin ();
	for (int i = 0; i <= dest && temp != _fBuff.end(); i++, temp++);
	temp--;
	if (temp != _fBuff.end()) {
		list <string> :: iterator next = temp;
		next++;
		if(next->length()) {
			string :: iterator stemp = temp->begin();
			for (int i = 0; i < temp->length() && (*stemp) != '\n'; i++, stemp++);		
			if ((*stemp) == '\n') {
				temp->erase(stemp);
				retval = 1;
			}
			(*temp) = (*temp) + (*next);
		}
		_fBuff.erase (next);
		return retval;
	}
	else beep();
	return retval;
}


buffer :: buffer () : _fBuff ()
{
	_fBuff.push_back("");
	_fBuff.begin()->push_back('\n');
	_fileNum = numOpenFiles;
}

buffer :: buffer (openFile & file, int & nRows) : _fBuff()
{
	initBuffer(file, nRows);
	_fileNum = numOpenFiles;
}

int buffer::initBuffer(openFile & file, int & nRows)
{
	fstream infile (file.fileName(), ios::in);
	mvwprintw (file.fileWin().outerWin(), 0 , file.fileWin().oCols() / 2 - strlen (file.fileName()) / 2 - 3, "[%s]", file.fileName());
	char temp;
	_fBuff.push_back("");
	_cRow = _fBuff.begin();
	int y = 0, x = 0;
	nRows = 0;
	while ((temp = infile.get()) != EOF) {
		if (temp != '\t' && temp != '\n' && isdisplay (temp)) {
			file.fileWin().putChar(temp, y, x, 0);
			_cRow->push_back(temp);
			x++;
		}
		else if (temp == '\t')
			for (int count = 1; count <= TABSZ; count++) {
				 file.fileWin().putChar(' ', y, x, 0);
				x++;
				_cRow->push_back(' ');
			}
		else if (temp == '\n') {
			nRows++;
			if (( x = file.fileWin().putChar(temp, y, x, 0)) == 2)
				;//file.updatePad();
			_cRow->push_back(temp);
			_fBuff.push_back("");
			_cRow++;
			y++;
			x = 0;
		}
	}
	_fBuff.erase(_cRow);
	if (!x && !y){
		file.status() = NEWFILE;
		nRows = 1;
		_fBuff.push_back("");
	}
	infile.close();
	return 1;
}

int buffer :: saveBuffer (char *name, int nLines)
{
	fstream outfile(name, ios::out);
	if(!outfile) {
		mesg("Permission denied :(");
		outfile.close();
		return -1;
	}
	list <string> :: iterator temp = _fBuff.begin();
	int i;
	for (i = 0; i <= nLines && temp != _fBuff.end(); i++, temp++) {
		if ((*temp)[temp->length() - 1] != '\n') {
			temp->push_back('\n');
		}
		outfile << (*temp);
	}
	outfile.close();
	return 1;
}

int buffer :: lastChar (int line)
{
	int i;
	list <string> :: iterator temp;
	for (temp = _fBuff.begin(), i = 0; i <= line && temp != _fBuff.end(); i++, temp++);
	temp--;
	return (*temp)[temp->length() - 1];
}

buffer :: ~buffer ()
{
	_fBuff.erase (_fBuff.begin(), _fBuff.end());
}

int buffer :: numChars (int line){
	int i;
	list<string>::iterator temp = _fBuff.begin();
	for (i = 0; i <= line && temp != _fBuff.end(); i++, temp++);
	temp--;
	if (temp != _fBuff.end()) return temp->length();
	else return 0;
}
			
						
int buffer :: insCh (int ch, int & nRows, int y, int x)
{
	
	_cRow = _fBuff.begin();
	for (int i = 0; i < y; i++)
		_cRow++;
	_cCol = _cRow->begin() + x;
	if (INS && ch != '\n') {
		_cRow->insert(_cCol, ch);
	}
	else if (ch != '\n'){
		_cRow->erase(_cCol);
		_cRow->insert(_cCol, ch);
	}	
	else if (ch == '\n') {
		string *temp = new string;
		temp->insert (temp->begin(), _cCol, _cRow->end());
		_cRow->erase (_cCol, _cRow->end());
		_cRow->push_back ('\n');
		if (temp->length()) 
			_fBuff.insert (++_cRow, *temp);
		else {
			_fBuff.insert (++_cRow, "");
		}
		nRows++;
		delete temp;
	}
	return 1;
}
	
int buffer :: delCh (int y, int x, int & ch) 
{
	list <string> :: iterator temp = _fBuff.begin();
	string :: iterator stemp;
	for (int i = 0; i <= y && temp != _fBuff.end(); i++, temp++);
	temp--;
	if ( (x < temp->length() - 1 || (*temp)[x] != '\n') && temp->length()) {
		stemp = temp->begin() + x;
		ch = *stemp;
		temp->erase(stemp);		
		return 1;
	}
	else {	
		stemp = temp->begin();
		if (temp->length() && (*temp)[x] == '\n') {
			stemp = stemp + x;
			temp->erase(stemp);
			ch = '\n';
		}
		list <string> :: iterator nextline = temp;
		nextline++;
		if (nextline != _fBuff.end())
		temp->insert(temp->end(), nextline->begin(), nextline->end());
		if (nextline != _fBuff.end()) _fBuff.erase(nextline);
		else return 3;
		return 2;
	}
	return 0;
}
	
int buffer :: delLn (int y, int & nRows)
{
	list <string> :: iterator temp = _fBuff.begin();
	for (int i = 0; i <= y && temp != _fBuff.end(); i++, temp++);
	temp--;
	_fBuff.erase (temp);
	--nRows;
	if (nRows == 0) {
		_fBuff.push_back("");
		_fBuff.begin()->push_back('\n');
		nRows++;
	}
	return 1;
}

int buffer :: find (string &pattern, int &y, int &x, int nRows, int beginY, int beginX)
{
	y = beginY, x = beginX;
	if (beginY == -1) y = 0;
	list <string> :: iterator temp = _fBuff.begin();
	int i = 0;
	while (i < y) {
		temp++;
		i++;
	}
	while (y < nRows) {
		x = temp->find(pattern.c_str());
		if (beginX == -1 && x != -1)
			return 1;
		if (x != -1  && beginY != -1 && x <= beginX && y == beginY) {
			x = temp->find (pattern.c_str(), beginX + 1);
		}
		if (x != -1)
			return 1;
		y++;
		temp++;
	}
	return 0;
}

int buffer :: replace (int y, int x, string & str, int n)
{
	list <string> :: iterator temp = _fBuff.begin();
	int i = 0;
	while (i < y) {
		i++;
		temp++;
	}
	temp->replace(x, n, str);
	return 1;
}


//---------------------------------------------------------------------------------------------

int window :: getChar ()
{
	return wgetch(_pad);
}


int window :: delCh (int y, int x, int ch, int key)
{
	if (key == BKSP)
		_cCol--;
	mvwdelch (_pad, _cRow, _cCol);
	return 1;
}

int window :: setXY (int y, int x){
	_cRow += y;
	_cCol += x;
	updateXY();
	return 1;
}


int window :: resize () 
{
	curs_set(0);
	RESIZE = 1;
	(*currentFile)->updateXY();
	int ch = wgetch(_pad);
	while (1){
		switch (ch) {
			case KEY_DOWN:
				_oRows++;
				break;
			case KEY_UP:
				_oRows--;
				break;
			case KEY_RIGHT:
				_oCols++;
				break;
			case KEY_LEFT:
				_oCols--;
				break;
			case 'F':
			case 'f':
				_oRows = LINES - 2;
				_oCols = COLS;
				break;
			default:
				RESIZE = 0;
				return 0;
		}
		if (_oRows < 3) {
			_oRows = 3;
			beep();
		}
		if (_oCols < 40) {
			_oCols = 40;
			beep();
		}
		if (_oRows > LINES - 2) {
			_oRows = LINES - 2;
			beep();
		}
		if (_oCols > COLS) {
			_oCols = COLS;
			beep();
		}
		_Y2 = _oRows - 1;
		_X2 = _oCols - 2;
		del_panel(_outerPanel);
		delwin (_outerWin);
		_outerWin = newwin (_oRows, _oCols, _beginY, _beginX);
		wbkgd(_outerWin, COLOR_PAIR (_pairnum));
		box (_outerWin, ACS_VLINE, ACS_HLINE);
		if (_oRows > 4) {
			mvwvline (_outerWin, 1, _oCols - 1, ACS_CKBOARD, _oRows - 2);
			mvwaddch (_outerWin, 1, _oCols - 1, ACS_UARROW);
			mvwaddch (_outerWin, _oRows - 2, _oCols - 1, ACS_DARROW);
		}
		else { 
			mvwaddch (_outerWin, 1, _oCols - 1, ACS_CKBOARD);
			if (_oRows > 3)
				mvwaddch (_outerWin, 2, _oCols - 1, ACS_CKBOARD);
		}
		_outerPanel = new_panel(_outerWin);
		mvwprintw (_outerWin, 0 , _oCols / 2 - strlen ((*currentFile)->fileName()) / 2 - 3, "[%s]", (*currentFile)->fileName());
		(*currentFile)->updateXY();
		ch = wgetch(_pad);
	}
	RESIZE = 0;
//	curs_set(1);
	return 1;
}

window :: window (int nRows, int nCols, int bgcolor, int fgcolor)
{
	_beginX = 0;
	_beginY = 1;
	_oRows = LINES - 2; // 15
	_oCols = COLS; // 60
	_outerWin = newwin(_oRows,_oCols, _beginY, _beginX);
	wmove(_outerWin, 1, 1);
	_outerPanel = new_panel (_outerWin);

	_Y1 = 2;
	_X1 = 1;
	_Y2 = _oRows - 1;
	_X2 = _oCols - 2;
	_pad = newpad(nRows, nCols);
	_bg = bgcolor;
	_fg = fgcolor;
	init_pair (++PAIRNUM, _fg, _bg);
	wbkgd(_outerWin, COLOR_PAIR(PAIRNUM));
	wbkgd(_pad, COLOR_PAIR(PAIRNUM));
	box(_outerWin, ACS_VLINE, ACS_HLINE);
	mvwaddch (_outerWin, 1, _oCols - 1, ACS_UARROW);
	mvwvline (_outerWin, 2, _oCols - 1, ACS_CKBOARD, _oRows - 3);
	mvwaddch (_outerWin, _oRows - 2, _oCols - 1, ACS_DARROW);
	_pairnum = PAIRNUM;
	_nRows = nRows;
	_nCols = nCols;
	_cRow = _cCol = 0;
	_cOrgY = 0;
	_cOrgX = 0;
	keypad (_pad, 1);
	scrollok(_pad, 1);
	wmove (_pad,_cRow,_cCol);
//	refreshWin();
//	refreshPad();
}

window :: ~window ()			// delwin() here
{
	del_panel(_outerPanel);
	delwin(_pad);
	delwin(_outerWin);
}

int window :: changeColor (int bgOrFg) 
{
	int nItems = 9;
	ITEM * colrItems[9];
	colrItems[0] = new_item ("  Black", "");
	colrItems[1] = new_item ("  White", "");
	colrItems[2] = new_item ("  Blue", "");
	colrItems[3] = new_item ("  Cyan", "");
	colrItems[4] = new_item ("  Red", "");
	colrItems[5] = new_item ("  Magenta    ", "");
	colrItems[6] = new_item ("  Green", "");
	colrItems[7] = new_item ("  Yellow", "");
	colrItems[8] = NULL;
	int ch, index, colr;
	update_panels();
	doupdate();
	(*currentFile)->updatePad();
	WINDOW *disp = newwin(10, 16, 7, 30);
	wbkgd(disp, COLOR_PAIR(_pairnum));
	wattron (disp, A_BOLD);
	if (bgOrFg == BG)
		mvwprintw(disp, 0, 3, "BACKGROUND");
	else 
		mvwprintw(disp, 0, 3, "FOREGROUND");
	wrefresh(disp);
	index =	displayMenu (colrItems, nItems, 0, 8, 30, ch);
	if (index == NOT_SEL) return 0;
	switch (index) {
		case 0:
			colr = COLOR_BLACK;
			break;
		case 1:
			colr = COLOR_WHITE;
			break;
		case 2:
			colr = COLOR_BLUE;
			break;
		case 3:
			colr = COLOR_CYAN;
			break;
		case 4:
			colr = COLOR_RED;
			break;
		case 5:
			colr = COLOR_MAGENTA;
			break;
		case 6:
			colr = COLOR_GREEN;
			break;
		case 7:
			colr = COLOR_YELLOW;
			break;
		default :
			return -1;
	}	
	delwin(disp);
	if (bgOrFg == BG)
		_bg = colr;
	else 
		_fg = colr;
	init_pair (_pairnum, _fg, _bg);
	wbkgd(_outerWin, COLOR_PAIR (_pairnum));
	wbkgd(_pad, COLOR_PAIR(_pairnum));
	updateXY();
	return 0;
}

int window :: refreshWin()
{
	curs_set(0);
	mvwprintw (_outerWin, _oRows - 1, _oCols - 30, "[ROW = %3d COL = %3d]", _cRow + 1, _cCol + 1);
	if (numOpenFiles > 1) 
		if ((*currentFile)->MARKSTRT)
			mvwprintw (_outerWin,_oRows - 1, _oCols - 5, "[M]" );
		else 
			mvwhline (_outerWin, _oRows - 1, _oCols - 5,  ACS_HLINE, 3);
	whline (_outerWin, ACS_HLINE, 1);
	whline (_outerWin, ACS_HLINE, 1);
	top_panel(_outerPanel);
	update_panels();
	doupdate();
//	curs_set(1);
	return 1;
}

void window :: updateXY ()
{
	refreshWin();
	refreshPad();
}

int window :: putChar (int ch, int y, int x, int nRowsBuf) 
{
	if (INS)			
		mvwinsch(_pad, _cRow, _cCol,(chtype) ch);
	else
		mvwaddch(_pad,_cRow, _cCol, (chtype) ch);
	_cCol++;
	if (ch == '\n'){		
		_cRow++;
		_cCol = 0;
		if (nRowsBuf >= _nRows - 2 || _cRow >= _nRows - 2) {
			delwin(_pad);
			_nRows = 2 * _nRows;
			_pad = newpad (_nRows, _nCols);
			wbkgd(_pad, COLOR_PAIR(PAIRNUM));
			keypad (_pad, 1);
			scrollok(_pad, 1);
			return 2;
		}
	}
	return (1);
}


int window :: refreshPad ()
{
	wmove(_pad, _cRow, _cCol);
	if (_cRow > (_cOrgY + _oRows - 3))
		 _cOrgY += 1;
	prefresh(_pad, _cOrgY, _cOrgX, _Y1, _X1, _Y2, _X2);
	return 1;
}

//---------------------------------------------------------------------------------------------


int main (int argc, char *argv[])
{
	extern list <openFile *> fileList;
	extern list <openFile *> :: iterator currentFile;
	extern int numOpenFiles;
	if (argc > 2) {
		cerr << "mD: confused by too many arguments, terminating program." << endl;
		exit (1);
	}
	initProg ();
	initMainWin ();
	if (argc == 2) {
		numOpenFiles++;
		fileList.push_front(new openFile(argv[1]));
		currentFile = fileList.begin();
		work(**currentFile);
	}
	else if (argc == 1) {
		wait();
	}
	terminateProg (0);
}

int wait ()
{
	int ch = wgetch(stdWin);
	while (1) {
		if (noFilesHandleCh(ch) == EXIT_CODE) break;
		ch = wgetch(stdWin);
	}
	return 1;
}

int noFilesHandleCh (int ch)
{
	if (ch == -1) return -1;
	switch (ch) {
		case KEY_ALT : 
			return (noFilesMenu());
		case KEY_RESIZE:
			beep();
			break;
		case KEY_INS:
			INS = !INS;
			if (INS)
				mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT");
			else 
				mvwprintw (stdWin, LINES - 1, COLS - 9, "      ");
			wrefresh(stdWin);
			wrefresh(mainWindow);
			break;		
		case KEY_F(1):
			newFile();
			break;
		case KEY_F(11):
			goShell();
			break;
		case KEY_F(3):
			openFil();
			break;
		case KEY_F(12):
			terminateProg(0);
		default :
			;	
	}
	top_panel(stdPanel);
	top_panel(mainPanel);
	update_panels();
	doupdate();
	return 1;
}

int noFilesMenu ()
{
	int ch = wgetch(stdWin);
	int off[15], retVal = 0;
	sw:
	switch (ch) {
		case 'f' :
		case 'F' :
			off[0] = 2;
			off[1] = 3;
			off[2] = 4;
			off[3] = 5;
			off[4] = 6;
			off[5] = 7;
			off[6] = 9;
			off[7] = 10;
			retVal = fileMenu(off, 8);
			break;
		case 'e' :
		case 'E' :
			for (int i = 0; i <= 12; i++)
				off[i] = i;
			retVal = editMenu(off, 13);
			break;
		case 's':
		case 'S':
			for (int i = 0; i <= 4; i++)
				off[i] = i;
			retVal = searchMenu (off, 5);
			break;
		case 'w' :
		case 'W' : 
			for (int i = 0; i <= 5; i++)
				off[i] = i;
			retVal = windMenu(off, 6);
			break;
		case 'h' :
		case 'H' :
			retVal = helpMenu(off, 0);
			break;
		case 'x' :
		case 'X' :
			terminateProg(0);
		default:
				;
	}
	if (retVal != 0) {
		ch = retVal;
		goto sw;
	}
	top_panel(stdPanel);
	top_panel(mainPanel);
	update_panels();
	doupdate();
	return 1;

}

void terminateProg (int code)
{
	delwin (mainWindow);
	delwin (stdWin);
	endwin();
	tcsetattr(0, TCSANOW, &init_settings);
	exit (code);
}

void initProg ()
{	
	tcgetattr(0, &init_settings);
	initscr(); 
	start_color();
	noecho();
	cbreak();
}


int displayMenu (ITEM **items, int nItems, int offItems, int beginY, int beginX, int & retch)
{
	curs_set(0);
	MENU * menu = new_menu (items);
	set_menu_fore(menu, A_BOLD);
	set_menu_mark (menu, "");
	set_menu_grey (menu, A_DIM);
	menu_opts_off (menu, O_NONCYCLIC);
	set_menu_spacing (menu, 5, 1, 1);
	WINDOW *fw;
	PANEL * fp;
	WINDOW *fsw;
	PANEL *fsp;
	int row, col;
	scale_menu (menu, &row, &col);
	fw = newwin (row+2, col+2, beginY, beginX);
	fp = new_panel(fw);
	top_panel(fp);
	wbkgd(fw, COLOR_PAIR(menucolr));
	
	fsw = derwin (fw, row, col, 1, 1);
	fsp = new_panel(fsw);
	top_panel (fsp);
	
	keypad (fsw, 1);
	box (fw, ACS_VLINE, ACS_HLINE);
	set_menu_win (menu, fw);
	set_menu_sub (menu, fsw);
	post_menu (menu);
	update_panels();
	doupdate();
	int ch, index = -1;
	while ((ch = wgetch(fsw)) != '\n') {	
		switch (ch) {
			case KEY_DOWN :
				if (nItems != offItems + 1) {
					menu_driver (menu, REQ_DOWN_ITEM);
					while ( item_opts(current_item(menu)) != O_SELECTABLE)
						menu_driver (menu, REQ_DOWN_ITEM);
				}
				break;
			case KEY_UP :
				if (nItems != offItems + 1) {
					menu_driver (menu, REQ_UP_ITEM);
					while ( item_opts(current_item(menu)) != O_SELECTABLE)
						menu_driver (menu, REQ_UP_ITEM);
				}
				break;
			case KEY_RIGHT :
				deleteMenu (nItems, menu, items, fw, fsw, fp, fsp);
				return NEXT_MENU;
			case KEY_LEFT :
				deleteMenu (nItems, menu, items, fw, fsw, fp, fsp);
				return PREV_MENU;
			default :;
		}
		if (isdisplay(ch)) {
			retch = ch;
			break;;
		}
		else if (ch == 27) {
			mvwprintw (stdWin, 0, 0, "%-s", " FILE     EDIT     SEARCH     WINDOW     HELP     ");
			break;
		}
	}
	if (ch == '\n' && nItems != offItems + 1) {
		ITEM *selected = current_item(menu);
		index = item_index(selected);
		deleteMenu (nItems, menu, items, fw, fsw, fp, fsp);
		return index;
	}
	deleteMenu (nItems, menu, items, fw, fsw, fp, fsp);
	return NOT_SEL;
}
	
int deleteMenu (int nItems, MENU * menu, ITEM ** items, WINDOW *fw, WINDOW *fsw, PANEL* fp, PANEL *fsp)
{
	unpost_menu(menu);
	free_menu(menu);
	for (int i = 0; i < nItems; i++)
		free_item (items[i]);
	del_panel(fp);
	del_panel(fsp);
	delwin (fsw);
	delwin (fw);
	mvwprintw (stdWin, 0, 0, "%-s", " FILE     EDIT	   SEARCH     WINDOW     HELP     ");	
	update_panels();
	doupdate();
	return 1;
}

int saveAll ()
{
	int i = 0;
	list <openFile *> :: iterator temp = fileList.begin();
	while (i < numOpenFiles) {
		(*temp)->saveFile();
		temp++;
		i++;
	}
	return 1;
}

int fileMenu (int *offItems, int numOffItems) 
{
	int nItems = 13;
	ITEM * fileItems[13];
	fileItems[0] = new_item ("New", " F1");
	fileItems[1] = new_item ("Open", " F3");
	fileItems[2] = new_item ("Save", " F2");
	fileItems[3] = new_item ("Save As", "");
	fileItems[4] = new_item ("Save All", "");
	fileItems[5] = new_item (" ", "");
	fileItems[6] = new_item ("File Info", "");
	fileItems[7] = new_item (" ", "");
	fileItems[8] = new_item ("Shell", "F11");
	fileItems[9] = new_item (" ", "");
	fileItems[10] = new_item ("Close", " F4");
	fileItems[11] = new_item ("Exit", "F12");
	fileItems[12] = NULL;
	wattron (stdWin, A_STANDOUT);
	mvwprintw (stdWin, 0, 0, "%-s", " FILE ");
	wattroff (stdWin, A_STANDOUT);
	for (int i = 0; i < numOffItems; i++)
		item_opts_off (fileItems [offItems[i]], O_SELECTABLE);
	int ch, index;
	char *s;
	string str;
	index = displayMenu (fileItems, nItems, numOffItems, 1, 0, ch);
	if (index == NOT_SEL) return 0;
 	switch (index) {
		case NEXT_MENU :
			return 'e';
		case PREV_MENU :
			return 'h';
		case 0:
			newFile();
			break;
		case 1 :
			openFil();
			break;
		case 2 :
			(*currentFile)->saveFile ();
			break;
		case 3:
			(*currentFile)->saveAs();
			break;
		case 4:
			saveAll();
			break;
		case 6:
			s = new char[200];
			sprintf(s, "  %s", (*currentFile)->fileName());
			str = string (s);
			str.push_back ('\n');
			sprintf(s, "  %d Line(s)", (*currentFile)->nRows());
			str += string (s);
			str.push_back('\n');
			sprintf(s, "  %d Words(s)", (*currentFile)->wordCnt());
			str += string (s);
			str.push_back ('\n');
			sprintf(s, "  %d Chars(s)", (*currentFile)->charCnt() - (*currentFile)->nRows());
			str += string (s);
			str.push_back ('\n');
			(*currentFile)->updateXY();
			prompt (str); 
			(*currentFile)->updateXY();
			delete []s;
			break;
		case 8:
			goShell();
			break;
		case 10 :
			(*currentFile)->closeFile();
			break;
		case 11:
			terminateProg(0);
			break;
		default :
			;
	}
	return 0;
}

int newFile()
{
	numOpenFiles++;	
	fileList.push_front(new openFile());
	currentFile = fileList.begin();
	mvwprintw ((*currentFile)->fileWin().outerWin(), 0, (*currentFile)->fileWin().oCols() / 2 - strlen ((*currentFile)->fileName()) / 2 - 3, "[%s]", (*currentFile)->fileName());
	if (INS)
		mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT");
	else 
		mvwprintw (stdWin, LINES - 1, COLS - 9, "      ");		
	(*currentFile)->updatePad();
	work(**currentFile);
	return 1;
}

int editMenu (int *offItems, int numOffItems)
{
	int nItems = 14;
	ITEM * editItems[14];

	editItems[0] = new_item ("Cut", "Alt X");
	editItems[1] = new_item ("Copy", "Alt C");
	editItems[2] = new_item ("Paste", "Alt V");
	editItems[3] = new_item ("Select All", "Alt A");
	editItems[4] = new_item ("Delete", "  DEL");
	editItems[5] = new_item (" ", "");
	editItems[6] = new_item ("Insert Word", "Alt I");
	editItems[7] = new_item ("Insert Line", "Alt L");
	editItems[8] = new_item ("Delete Word", "Alt T");
	editItems[9] = new_item ("Delete Line", "Alt Y");
	editItems[10] = new_item (" ", " ");
	editItems[11] = new_item ("Toggle Insert", "  INS");
	editItems[12] = new_item ("Mark Mode", "Alt M");
	editItems[13] = NULL;
	for (int i = 0; i < numOffItems; i++)
		item_opts_off (editItems [offItems[i]], O_SELECTABLE);
	wattron (stdWin, A_STANDOUT);
	mvwprintw (stdWin, 0, 9, "%-s", " EDIT ");	
	wattroff (stdWin, A_STANDOUT);

	int ch, index;
	index =	displayMenu (editItems, nItems, numOffItems, 1, 9, ch);
	int cnt;
	if (index == NOT_SEL) return 0;
	switch (index) {
		case NEXT_MENU :
			return 's';
		case PREV_MENU :
			return 'f';
		case 0:
			(*currentFile)->cut();
			break;
		case 1:
			(*currentFile)->copy();
			break;
		case 2:
			(*currentFile)->paste();
			break;
		case 3:
			(*currentFile)->selectAll();
			break;
		case 4:
			(*currentFile)->delSelection();
			break;
		case 6:
			(*currentFile)->insWord();
			break;
		case 7:
			(*currentFile)->insLn();
			break;
		case 8:
			(*currentFile)->delWord();
			break;
		case 9:
			(*currentFile)->delLn();
			break;
		case 11:
			INS = !INS;
			if (INS)
				mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT");
			else 
				mvwprintw (stdWin, LINES - 1, COLS - 9, "      ");
			break;
		case 12:
			(*currentFile)->beginMark();
			(*currentFile)->MARKSTRT = 1;
			break;
		default :
			;
	}
	return 0;
}

int helpMenu (int *offItems, int numOffItems) {
	int nItems = 3;
	ITEM * helpItems[3];
	helpItems[0] = new_item (" Help Topics ", "");
	helpItems[1] = new_item (" About mD", "");
	helpItems[2] = NULL;
	int ch, index;
	wattron (stdWin, A_STANDOUT);
	mvwprintw (stdWin, 0, 41, "%-s", " HELP ");
	wattroff (stdWin, A_STANDOUT);

	index = displayMenu (helpItems, nItems, numOffItems, 1, 41, ch);
 	if (index == NOT_SEL) return 0;
	switch (index) {
		case NEXT_MENU :
			return 'f';
		case PREV_MENU :
			return 'w';
		case 0:
			helpfile();
			break;
		case 1:
			about ();
			break;
		default :
			;
	}
	return 0;
}

int searchMenu (int *offItems, int numOffItems)
{
	int nItems = 6;
	ITEM * searchItems[6];
	searchItems[0] = new_item ("Find", "   F5");
	searchItems[1] = new_item ("Find Again", "   F6");
	searchItems[2] = new_item ("Replace","   F7");
	searchItems[3] = new_item (" ", "");
	searchItems[4] = new_item ("Goto Line", "Alt G");
	searchItems[5] = NULL;
	for (int i = 0; i < numOffItems; i++)
		item_opts_off (searchItems [offItems[i]], O_SELECTABLE);
	int ch, index;
	char *s;
	wattron (stdWin, A_STANDOUT);
	mvwprintw (stdWin, 0, 18, "%-s", " SEARCH ");	
	wattroff (stdWin, A_STANDOUT);

	index =	displayMenu (searchItems, nItems, numOffItems, 1, 18, ch);
 	if (index == NOT_SEL) return 0;
	switch (index) {
		case NEXT_MENU :
			return 'w';
		case PREV_MENU :
			return 'e';
		case 0:
			(*currentFile)->find();
			break;
		case 1:
			(*currentFile)->findAgain();
			break;
		case 2:
			(*currentFile)->replace();;
			break;
		case 4:
			(*currentFile)->gotoLine();
			break;
		default:
			;
	}	
	return 0;
}

int cycleWindows(int dirn) 
{ 
	if (numOpenFiles <= 1) {
		beep();
		return 1;
	}
	list <openFile*> :: iterator temp = fileList.begin();
	if (dirn == PREV) {
		fileList.push_back(*temp);
		fileList.pop_front();
	}	
	else  {
		int i = 1;
		while (i++ != numOpenFiles)
			temp++;
		fileList.push_front (*temp);
		fileList.pop_back();
	}
	currentFile = fileList.begin();
	(*currentFile)->updateXY();
	work (**currentFile);
	return 1;
}

int windMenu (int *offItems, int numOffItems) 
{
	int nItems = 7;
	ITEM * windItems[7];
	windItems[0] = new_item ("Resize Window", " F8");
	windItems[1] = new_item ("Previous Windows", " F9");
	windItems[2] = new_item ("Next Window", "F10");
	windItems[3] = new_item (" "," ");
	windItems[4] = new_item ("BG Color", "");
	windItems[5] = new_item ("Text Color", "");
	windItems[6] = NULL;
	for (int i = 0; i < numOffItems; i++)
		item_opts_off (windItems [offItems[i]], O_SELECTABLE);
	wattron (stdWin, A_STANDOUT);
	mvwprintw (stdWin, 0, 29, "%-s", " WINDOW ");
	wattroff (stdWin, A_STANDOUT);

	int ch, index;
	char *s;
	index =	displayMenu (windItems, nItems, numOffItems, 1, 29, ch);
 	if (index == NOT_SEL) return 0;
	switch (index) {
		case NEXT_MENU :
			return 'h';
		case PREV_MENU :
			return 's';
		case 0:
			(*currentFile)->fileWin().resize();
			break;
		case 1:
			cycleWindows(PREV);
			break;
		case  2:
			cycleWindows(NEXT);
			break;
		case 4:
			(*currentFile)->fileWin().changeColor(BG);
			break;
		case 5:
			(*currentFile)->fileWin().changeColor(FG);
			break;
		default:
			;
	}	
	return 0;
}


int openFil ()
{
	if (numOpenFiles)
		(*currentFile)->updateXY();
	int flag = OK;
	string fileName = browse ("Open File", flag);
//	string fileName = getIpStr ("FileName :", flag);
	if (flag == CANCEL){
		if (numOpenFiles)
			(*currentFile)->updateXY();
		return 0;
	}
	numOpenFiles++;	
	fileList.push_front(new openFile(fileName.c_str()));
	currentFile = fileList.begin();
	mvwprintw ((*currentFile)->fileWin().outerWin(), 0 , (*currentFile)->fileWin().oCols() / 2 - strlen ((*currentFile)->fileName()) / 2 - 3, "[%s]", (*currentFile)->fileName());
	if (INS)
		mvwprintw (stdWin, LINES - 1,COLS - 9, "INSERT");
	else 
		mvwprintw (stdWin, LINES - 1, COLS - 9, "      ");		
	(*currentFile)->updatePad();
	work(**currentFile);
	return 1;
}

int goShell ()
{
	tcsetattr(0, TCSANOW, &init_settings);
	endwin();
	system("clear");
	system("echo Type exit to return to mD");
	system("bash");
	system("clear");
	refresh();
	noecho();
	cbreak();
	return 1;
}

int initMainWin ()
{
	curs_set (0);
	init_pair (++PAIRNUM, COLOR_CYAN, COLOR_BLACK);
	stdWin = newwin (LINES, COLS, 0, 0);
	stdPanel = new_panel(stdWin);
	wbkgd(stdWin, COLOR_PAIR(PAIRNUM));
	wattron (stdWin, A_BOLD);
	stdcolr = PAIRNUM;
	mainWindow = newwin (LINES - 2, COLS, 1, 0);
	mainPanel = new_panel(mainWindow);
	init_pair (++PAIRNUM, COLOR_BLACK, COLOR_CYAN);
	wbkgd (mainWindow, chtype (ACS_BOARD) | COLOR_PAIR(PAIRNUM));
	maincolr = PAIRNUM;
	keypad(stdWin, 1);
	keypad(mainWindow, 1);
	wrefresh(stdWin);
	wrefresh(mainWindow);
	diacolr = maincolr;
	butcolr = stdcolr;
	init_pair (++PAIRNUM, COLOR_WHITE, COLOR_BLACK);
	inpcolr = PAIRNUM;
	mvwprintw (stdWin, 0, 0, "%-s", " FILE     EDIT	   SEARCH     WINDOW     HELP     ");	
	mvwprintw (stdWin, LINES - 1, 1,  "%-*s", COLS - 12,  "F1 - New | F2 - Save | F3 - Open | F11 - Shell | F12 - Exit");
	if (INS)
			mvwprintw(stdWin, LINES - 1, COLS - 9, "INSERT");
	WINDOW *logoshadow = newwin (10, 27, LINES / 2 - 5, COLS / 2 - 14);
	PANEL * lsp = new_panel (logoshadow);
	top_panel (lsp);
	WINDOW *logowin = newwin (10, 27, LINES / 2 - 6, COLS / 2 - 15);
	PANEL * lp = new_panel (logowin);
	top_panel (lp);
	init_pair (++PAIRNUM, COLOR_BLACK, COLOR_CYAN);
	wbkgd (logowin, COLOR_PAIR (PAIRNUM));
	wbkgd (logoshadow, ACS_BOARD | COLOR_PAIR (butcolr));
	mvwprintw (logowin, 0, 1, "                _____");
	mvwprintw (logowin, 1, 1, "               |     \\ ");	
	mvwprintw (logowin, 2, 1, "               |      \\ ");
	mvwprintw (logowin, 3, 1, "  _____  _____ |  | \\  \\ ");
	mvwprintw (logowin, 4, 1, " /     \\/     \\|  |  \\  \\ ");
	mvwprintw (logowin, 5, 1, "|   /\\    /\\   |  |  /  / ");
	mvwprintw (logowin, 6, 1, "|  |  |  |  |  |  |_/  / ");
	mvwprintw (logowin, 7, 1, "|  |  |  |  |  |      / ");
	mvwprintw (logowin, 8, 1, "|__|  |__|  |__|_____/ ");
	mvwprintw (logowin, 9, 1,"       miniDrafter");
	update_panels();
	doupdate();
	wgetch (logoshadow);
	del_panel (lsp);
	del_panel (lp);
	delwin (logoshadow);
	delwin (logowin);
	update_panels();
	doupdate();
	init_pair (PAIRNUM, COLOR_CYAN, COLOR_BLACK);
	menucolr = PAIRNUM;
	return (1);
}


string getIpStr (const char* msg, int &flag, const char *def)
{
	string str;
	WINDOW *mesgWin = newwin (7, 40, LINES / 2 - 5, COLS / 2 - 25);
	PANEL *mesgPan = new_panel(mesgWin);
	wbkgd (mesgWin, COLOR_PAIR(diacolr));
	box (mesgWin, ACS_VLINE, ACS_HLINE);
	int len = strlen (msg);
	mvwprintw(mesgWin, 1, 3, "%s", msg);
	WINDOW *ipWin = derwin (mesgWin, 1, 34, 3, 3);
	WINDOW *ok = derwin (mesgWin, 1, 10, 5, 3);
	WINDOW *cancel = derwin (mesgWin, 1, 10, 5, 27);
	keypad(ipWin, 1);
	wbkgd (ipWin, COLOR_PAIR(inpcolr));
	if (def != 0) {
		wprintw (ipWin, (char *)def);
		str = string (def);
	}
	wbkgd (ok, COLOR_PAIR(butcolr));
	wbkgd (cancel, COLOR_PAIR(butcolr));	
	wattron (ok, A_BOLD);
	mvwprintw(ok, 0, 0, ">   OK   <");
	mvwprintw(cancel, 0, 0, "  CANCEL  ");	
	top_panel(mesgPan);
	update_panels();
	doupdate();
	wmove(ipWin, 0, 0);
	wrefresh (ipWin);
	wrefresh(ok);
	wrefresh(cancel);
	int retVal = OK, flg = 0;
	int ch = mvwgetch (ipWin, 0, 0);
	while (ch != '\n' && ch != 27) {
		if (isdisplay (ch)) {
			if (def != 0 && flg == 0) {
				str.erase (str.begin(), str.end());
				wprintw(ipWin, "				  ");
				wmove (ipWin, 0, 0);
				flg = 1;
			}
			waddch (ipWin, ch);
			str.push_back (ch);
		}
		else
			switch (ch) {
				case KEY_RIGHT:
					wattroff (ok, A_BOLD);
					mvwprintw(ok, 0, 0, "    OK    ");
					wrefresh(ok);
					wattron (cancel, A_BOLD);
					mvwprintw(cancel, 0, 0, "> CANCEL <");	
					wrefresh(cancel);
					retVal = CANCEL;
					break;
				case KEY_LEFT:
					wattron (ok, A_BOLD);
					mvwprintw(ok, 0, 0, ">   OK   <");
					wattroff(cancel, A_BOLD);
					mvwprintw(cancel, 0, 0, "  CANCEL  ");	
					wrefresh(cancel);
					wrefresh(ok);
					retVal = OK;
					break;
				case KEY_BACKSPACE: 
					if (str.length()) {
						str.erase(str.begin() + str.length() - 1);
						wmove(ipWin, 0, 0);
						mvwprintw(ipWin, 0, 0, "%-*s ", str.length() + 1, (char*) str.c_str());
						break;
					}
			}
		ch = mvwgetch(ipWin, 0, str.length());
		if (ch == 27) {
			retVal = CANCEL;
			break;
		}
	}
	del_panel(mesgPan);
	delwin (ipWin);
	delwin(ok);
	delwin(cancel);
	delwin(mesgWin);
	update_panels();
	doupdate();
	noecho();
	if (str.length())
		flag = retVal;
	else flag = CANCEL;
	return str; 
}

int helpfile()
{
	curs_set(0);
	WINDOW *mesgWin = newwin (18, 60, LINES / 2 - 8 , COLS / 2 - 29);
	PANEL *mesgPan = new_panel(mesgWin);
	wbkgd (mesgWin, COLOR_PAIR(butcolr) | ACS_BOARD);
	WINDOW *mesgSub = newwin(18, 60, LINES / 2 - 9, COLS / 2 - 30);
	PANEL *mesgSPan = new_panel (mesgSub);
	box (mesgSub, ACS_VLINE, ACS_HLINE);
	wbkgd (mesgSub, COLOR_PAIR (diacolr));	
	mvwprintw(mesgSub, 1, 18, "HOT keys and Commands");
	mvwhline (mesgSub, 2, 1, ACS_HLINE, 58);
	mvwprintw(mesgSub, 3, 2,  "Alt M       Begin / End Mark      F1    New");	
	mvwprintw(mesgSub, 4, 2,  "Alt X       Cut                   F2    Save");
	mvwprintw(mesgSub, 5, 2,  "Alt C       Copy                  F3    Open");
	mvwprintw(mesgSub, 6, 2,  "Alt V       Paste                 F4    Close");
	mvwprintw(mesgSub, 7, 2,  "Alt Y       Delete Line           F5    Find");
	mvwprintw(mesgSub, 8, 2,  "Alt T       Delete Word           F6    Find Again");	
	mvwprintw(mesgSub, 9, 2,  "Alt Y       Insert Line           F7    Replace");
	mvwprintw(mesgSub, 10, 2,  "Alt I       Insert Word           F8    Resize Window");
	mvwprintw(mesgSub, 11, 2, "Alt F       File Menu             F9    Next Window");
	mvwprintw(mesgSub, 12, 2, "Alt E       Edit Menu             F10   Previous Window");
	mvwprintw(mesgSub, 13, 2, "Alt S       Search Menu           F11   OS Shell");
	mvwprintw(mesgSub, 14, 2, "Alt W       Window Menu           F12   Quit");
	mvwprintw(mesgSub, 15, 2, "Alt H       Help Menu             DEL   Delete Selection");
	mvwprintw(mesgSub, 16, 2, "Alt G       Goto Line             INS   Toggle Insert");
	top_panel (mesgPan);
	top_panel (mesgSPan);
	update_panels();
	doupdate();
	wgetch (mesgSub);
	del_panel(mesgPan);
	del_panel(mesgSPan);
	delwin(mesgSub);
	delwin(mesgWin);
	return 1;
}

int about ()
{
	curs_set(0);
	WINDOW *mesgWin = newwin (8, 50, LINES / 2 - 4, COLS / 2 - 24);
	PANEL *mesgPan = new_panel(mesgWin);
	wbkgd (mesgWin, COLOR_PAIR (butcolr) | ACS_BOARD);
	WINDOW *mesgSub = newwin (8, 50, LINES / 2 - 5, COLS / 2 - 25);
	PANEL *mesgSPan = new_panel (mesgSub);
	box (mesgSub, ACS_VLINE, ACS_HLINE);
	wbkgd (mesgSub, COLOR_PAIR (diacolr));
	mvwprintw(mesgSub, 1, 2, "miniDrafter (mD) is a small and light weight ");
	mvwprintw(mesgSub, 2, 2, "ncurses based text editor.");	
	mvwprintw(mesgSub, 4, 2, "Authors :");
	mvwprintw(mesgSub, 5, 5, "> Krishna Jaidev S ");
	mvwprintw(mesgSub, 6, 5, "          - mail@jaidev.port5.com");
	top_panel (mesgPan);
	top_panel (mesgSPan);
	update_panels();
	doupdate();
	wgetch (mesgSub);
	del_panel(mesgPan);
	del_panel(mesgSPan);
	delwin(mesgSub);
	delwin(mesgWin);
	return 1;
}



string browse (string s, int & flg)
{
	curs_set(0);
	string ret;
	flg = CANCEL;
	char direct[100], pwd[100];
	vector <char *> files;
	vector <char *> filesCopy;
	vector <char *> directories;
	vector <char *> dirsCopy;
	ITEM * brows_files_items[1000];
	ITEM * brows_dirs_items[100];
	int y = LINES / 2 - 8;
	int x = COLS / 2 - 26;
	int flag = FILES;
	
	WINDOW *browse = newwin (17, 44, y, x);
	mvwhline (browse, 2, 0, ACS_HLINE, 44);
	WINDOW *sel = derwin (browse, 1, 11, 13, 2);
	WINDOW *ok = derwin (browse, 1, 10, 15, 5);
	WINDOW *cancel = derwin (browse, 1, 10, 15, 29);
	WINDOW *ipWin = derwin (browse, 1, 25, 13, 15);
	WINDOW *fls = derwin (browse, 1, 5, 3, 9);
	WINDOW *drs = derwin (browse, 1, 11, 3, 29);
	mvwprintw(sel, 0, 0, "Selection :");
	wbkgd (ipWin, COLOR_PAIR (inpcolr));
	wbkgd (browse, COLOR_PAIR (diacolr));
	wbkgd (ok, COLOR_PAIR(butcolr));
	wbkgd (sel, COLOR_PAIR (diacolr));
	wbkgd (fls, COLOR_PAIR(butcolr));
	wbkgd (drs, COLOR_PAIR(butcolr));
	wbkgd (cancel, COLOR_PAIR(butcolr));
	keypad(ok, 1);
	keypad(cancel, 1);
	keypad(ipWin, 1);
	wattron (fls, A_BOLD);
	mvwprintw (fls, 0, 0, "FILES");
	wattroff (fls, A_BOLD);
	mvwprintw (drs, 0, 0, "DIRECTORIES");
	mvwprintw(ok, 0, 0, "    OK    ");
	mvwprintw(cancel, 0, 0, "  CANCEL  ");	
	
	mvwprintw(browse, 1, 1, " %s...", s.c_str());
	box (browse, ACS_VLINE, ACS_HLINE);
	WINDOW *fw = newwin (8, 20, y + 4, x + 2);
	WINDOW *fs = derwin (fw, 8, 20, 0, 0);
	WINDOW *dw = newwin (8, 15, y + 4, x + 27);
	WINDOW *ds = derwin (dw, 8, 15, 0, 0);
	keypad(ds, 1);
	keypad(fs, 1);
	
	struct dirent *contents;
	
	int i = 0, numf = 0, numd = 0, j = 0, ch;
	MENU *brows_files_menu, *brows_dirs_menu;	
	DIR *cwd, *temp;
	struct stat statbuf;
	chdir (getenv ("HOME"));
dirhere:
	getcwd (direct, 100);
	if ((cwd = opendir (direct)) == NULL) {
		mesg ("Permission denied :-(");
		chdir (getenv("HOME"));
		goto dirhere;
	}
	i = 0;
	j = 0;
	numf = numd = 0;
	contents = readdir(cwd);
	
	while (1) {
		if (contents != NULL) {
			lstat (contents->d_name, &statbuf);
			if (!S_ISDIR(statbuf.st_mode)) {
				files.push_back (new char[sizeof(contents->d_name)+1]);
				filesCopy.push_back (new char[sizeof (contents->d_name) + 1]);
				sprintf(files[i], "%s", contents->d_name);
				sprintf(filesCopy[i++], "%s", contents->d_name);
			}
			else {
				directories.push_back (new char[sizeof (contents->d_name) + 1]);
				dirsCopy.push_back (new char[sizeof (contents->d_name) + 1]);
				sprintf (directories[j], "%s", contents->d_name);
				sprintf(dirsCopy[j++], "%s/", contents->d_name);
			}
			contents = readdir(cwd);
		}
		else
			break;
	}
	numf = i;
	numd = j;
	files[i] = NULL;
	filesCopy[i] = NULL;
	directories [j] = NULL;
	dirsCopy[j] = NULL;
	for (i = 0; filesCopy[i] != NULL; i++) {
		if (strlen (filesCopy[i]) > 20) {
			filesCopy[i][18] = '.';
			filesCopy[i][19] = '.';
			filesCopy[i][20] = '\0';
		}
		brows_files_items[i] = new_item (filesCopy[i], NULL);
	}
	brows_files_items[i] = NULL;
	for (j = 0; dirsCopy[j] != NULL; j++) {
		if (strlen (dirsCopy[j]) > 15) {
			dirsCopy[j][12] = '.';
			dirsCopy[j][13] = '.';
			dirsCopy[j][14] = '/';
			dirsCopy[j][15] = '\0';
		}
		brows_dirs_items[j] = new_item (dirsCopy[j], NULL);
	}
	brows_dirs_items[j] = NULL;

	brows_dirs_menu = new_menu (brows_dirs_items);
	brows_files_menu = new_menu (brows_files_items);
	set_menu_mark (brows_files_menu, "");
	set_menu_mark (brows_dirs_menu, "");
	set_menu_format (brows_files_menu, 8, 1);
	set_menu_format (brows_dirs_menu, 8, 1);
	
	set_menu_win (brows_files_menu, fw);
	set_menu_sub (brows_files_menu, fs);
	set_menu_win (brows_dirs_menu, dw);
	set_menu_sub (brows_dirs_menu ,ds);
	wrefresh (browse);
	wrefresh (sel);
	wrefresh(ok);
	wrefresh(cancel);
	wrefresh (fls);
	wrefresh (drs);
	wrefresh(dw);
	wrefresh(fw);
	post_menu(brows_files_menu); 
	post_menu(brows_dirs_menu);
	wrefresh(ds);
	wrefresh(fs);
	if (numf) {
		mvwprintw(ipWin, 0, 0, "%-20s", files[item_index(current_item(brows_files_menu))]);
		wmove (ipWin, 0, strlen (files[item_index(current_item(brows_files_menu))]));
	}
	wrefresh (ipWin);
	if (numf) {
		ch = browse_driver (brows_files_menu, fs, ipWin, filesCopy);	
		flag = FILES;
	}
	else {
		wattron (drs, A_BOLD);
		mvwprintw (drs, 0, 0, "DIRECTORIES");
		wattroff (drs, A_BOLD);
		wrefresh(drs);
		mvwprintw (fls, 0,0,"FILES");
		wrefresh(fls);
		ch = browse_driver (brows_dirs_menu, ds, ipWin, dirsCopy);
		flag = DIRS;
	}
	while (ch != 27) {
		if (ch == '\n' && flag == DIRS) {
			chdir (directories[item_index(current_item(brows_dirs_menu))]);
			unpost_menu(brows_files_menu);
			free_menu (brows_files_menu);
			for (i = 0; i < numf; i++)
				free_item (brows_files_items[i]);
			unpost_menu(brows_dirs_menu);
			free_menu (brows_dirs_menu);
			for (i = 0; i < numd; i++)
				free_item (brows_dirs_items[i]);
			files.erase (files.begin(), files.end());
			dirsCopy.erase (dirsCopy.begin(), dirsCopy.end());
			filesCopy.erase (filesCopy.begin(), filesCopy.end());
			directories.erase (directories.begin(), directories.end());
			flag = FILES;
			wattron (fls, A_BOLD);
			mvwprintw (fls, 0,0,"FILES");
			wattroff (fls, A_BOLD);
			mvwprintw (drs, 0, 0, "DIRECTORIES");
			goto dirhere;
		}
		else if (ch == '\n') {
			if (flag != CANBUT)
				flg = OK;
			else break;
			if (!ret.length()) {
				ret = string (direct);
				ret += '/';
				ret += files[item_index(current_item(brows_files_menu))];
			}
			else if (ret[0] != '/')
				ret = string (direct) + '/' + ret;
			ret.push_back ('\0');
			closedir (cwd);
			break;
		}
		else if (isdisplay(ch) && ch != '\t' && flag == INPUTFLD) {
			ret.push_back (ch);
			mvwprintw(ipWin, 0, 0, "%s", ret.c_str());
			wclrtoeol (ipWin);
		}
		else if (ch == KEY_BACKSPACE && flag == INPUTFLD && ret.length()) {
			ret.erase(ret.begin() + ret.length() - 1);
			mvwprintw(ipWin, 0, 0, "%-25s", ret.c_str());
			wclrtoeol (ipWin);
		}			
		flg = OK;
		switch (ch) {
			case KEY_TAB:
				flag++;
				if (flag > CANBUT)
					flag = FILES;
				break;
			case KEY_RIGHT :
				flag ++;
				if (flag > CANBUT)	
					flag = FILES;
				break;
			case KEY_LEFT:
				flag --;
				if (flag < FILES)
					flag = CANBUT;
				break;
			default :;
		}
		if (flag == INPUTFLD)
			curs_set(1);
		else 
			curs_set (0);
		mvwprintw (fls, 0,0,"FILES");
		mvwprintw (drs, 0, 0, "DIRECTORIES");
		if (flag != DIRS)
			wrefresh(drs);
		if (flag != FILES)
			wrefresh(fls);
		if (flag == FILES || flag == DIRS) {
			ret.erase (ret.begin(), ret.end());
		}
		if (flag != INPUTFLD) {
			mvwprintw(sel, 0, 0, "Selection :");
			wrefresh (sel);
		}
		switch (flag) {
			case FILES:
				mvwprintw(cancel, 0, 0, "  CANCEL  ");
				wrefresh(cancel);
				if (numf) {
					wattron (fls, A_BOLD);
					mvwprintw (fls, 0,0,"FILES");
					wattroff (fls, A_BOLD);
					wrefresh(fls);
					ch = browse_driver (brows_files_menu, fs, ipWin, filesCopy);	
				}
				else {
					wattron (drs, A_BOLD);
					mvwprintw (drs, 0, 0, "DIRECTORIES");
					wattroff (drs, A_BOLD);
					wrefresh(drs);
					ch = browse_driver (brows_dirs_menu, ds, ipWin, dirsCopy);
					flag = DIRS;
				}
				break;
			case DIRS:
				wattron (drs, A_BOLD);
				mvwprintw (drs, 0, 0, "DIRECTORIES");
				wattroff (drs, A_BOLD);
				wrefresh(drs);
				ch = browse_driver (brows_dirs_menu, ds, ipWin, dirsCopy);
				break;
			case INPUTFLD:
				mvwprintw(ok, 0, 0, "    OK    ");
				wrefresh(ok);
				wattron (sel, A_UNDERLINE);
				mvwprintw(sel, 0, 0, "Selection :");
				wattroff (sel, A_UNDERLINE);
				wrefresh(sel);
				wrefresh (ipWin);
				ch = mvwgetch (ipWin, 0, ret.length());
				break;
			case OKBUT:
				wattron (ok, A_BOLD);
				mvwprintw(ok, 0, 0, ">   OK   <");
				wattroff (ok, A_BOLD);
				mvwprintw(cancel, 0, 0, "  CANCEL  ");	
				wrefresh(cancel);
				flg = OK;
				ch = mvwgetch(ok, 0, 1);
				break;
			case CANBUT:
				wattron (cancel, A_BOLD);
				mvwprintw(cancel, 0, 0, "> CANCEL <");	
				wattroff (cancel, A_BOLD);
				mvwprintw(ok, 0, 0, "    OK    ");
				wrefresh(ok);
				flg = CANCEL;
				ch = mvwgetch(cancel, 0, 1);
				break;
		} // end switch (flag)
	}//end while 
	
	unpost_menu(brows_files_menu);
	unpost_menu(brows_dirs_menu);	
	free_menu (brows_files_menu);
	free_menu (brows_dirs_menu);
	for (i = 0; i < numf; i++)
		free_item (brows_files_items[i]);
	for (i = 0; i < numd; i++)
		free_item (brows_dirs_items[i]);
	files.erase (files.begin(), files.end());
	filesCopy.erase (filesCopy.begin(), filesCopy.end());
	dirsCopy.erase (dirsCopy.begin(), dirsCopy.end());
	directories.erase (directories.begin(), directories.end());
	delwin (sel);
	delwin (fs);
	delwin (fw);
	delwin (ds);
	delwin (dw);
	delwin (ipWin);
	delwin (ok);
	delwin (cancel);
	delwin (browse);
	if (ch == 27)
		flg = CANCEL;
	return ret;
}

int browse_driver (MENU * menu, WINDOW *sub, WINDOW *ipWin, vector <char *> flist)
{
	int ch;
	mvwprintw(ipWin, 0, 0, "%-25s", flist[item_index(current_item(menu))]);
	wrefresh (ipWin);
	while ((ch = wgetch(sub)) == KEY_DOWN || ch == KEY_UP) {
		switch (ch) {
			case KEY_DOWN :
				menu_driver (menu, REQ_DOWN_ITEM);
				break;
			case KEY_UP :
				menu_driver (menu, REQ_UP_ITEM);
				break;
			default:;
		}
		mvwprintw(ipWin, 0, 0, "%-25s", flist[item_index(current_item(menu))]);
		wmove (ipWin, 0, strlen (flist[item_index(current_item(menu))]));
		wrefresh (ipWin);
	}
	return ch;
}


int prompt (string &mesg)
{
	curs_set (0);
	WINDOW *outer = newwin (8, 40, LINES / 2 - 5, COLS / 2 - 25);
	WINDOW *mesgWin = derwin (outer, 4, 36, 1, 2);
	PANEL *mesgPan = new_panel(outer);
	wbkgd (mesgWin, COLOR_PAIR(diacolr));
	wbkgd (outer, COLOR_PAIR(diacolr));
	box (outer, ACS_VLINE, ACS_HLINE);
	mvwprintw(mesgWin, 0, 0, "%s", mesg.c_str());
	WINDOW *ok = derwin (outer, 1, 10, 6, 3);
	WINDOW *cancel = derwin (outer, 1, 10, 6, 27);
	keypad(ok, 1);
	keypad(cancel, 1);
	wbkgd (ok, COLOR_PAIR(butcolr));
	wbkgd (cancel, COLOR_PAIR(butcolr));	
	wattron (ok, A_BOLD);
	mvwprintw(ok, 0, 0, ">   OK   <");
	mvwprintw(cancel, 0, 0, "  CANCEL  ");	
	top_panel(mesgPan);
	update_panels();
	doupdate();
	wrefresh(ok);
	wrefresh(cancel);
	int retVal = OK;
	int ch = mvwgetch (ok, 0, 1);
	while (ch != '\n' && ch != 27) {
		switch (ch) {
			case KEY_RIGHT:
				wattroff (ok, A_BOLD);
				mvwprintw(ok, 0, 0, "    OK    ");
				wrefresh(ok);
				wattron (cancel, A_BOLD);
				mvwprintw(cancel, 0, 0, "> CANCEL <");	
				ch = mvwgetch (cancel, 0, 1);
				retVal = CANCEL;
				break;
			case KEY_LEFT:
				wattron (ok, A_BOLD);
				mvwprintw(ok, 0, 0, ">   OK   <");
				wattroff(cancel, A_BOLD);
				mvwprintw(cancel, 0, 0, "  CANCEL  ");	
				wrefresh(cancel);
				ch = mvwgetch (ok, 0, 1);
				retVal = OK;
				break;
			default:
				if (ch != '\n' && ch != 27)
					if (retVal == OK)
						ch = mvwgetch(ok, 0, 1);
					else 
						ch = mvwgetch(cancel, 0, 1);
				
		}
	}
	del_panel(mesgPan);
	delwin(ok);
	delwin(cancel);
	delwin(outer);
	delwin(mesgWin);
	if (ch == 27) 
		retVal = CANCEL;
//	curs_set (1);
	return retVal;
}

int mesg (const char *msg)
{
	update_panels();
	doupdate();
	curs_set (0);
	if (numOpenFiles)
		(*currentFile)->updateXY();
	WINDOW *mesgWin = newwin (7, 40, LINES / 2 - 5, COLS / 2 - 20);
	wbkgd (mesgWin, COLOR_PAIR(diacolr));
	box (mesgWin, ACS_VLINE, ACS_HLINE);
	int len = strlen (msg);
	mvwprintw(mesgWin, 3, 20 - len / 2, "%s", msg);
	wgetch(mesgWin);
	delwin(mesgWin);
//	curs_set (1);
	return 1;
}


int isdisplay (int ch)
{
	return ((ch >= ' ' && ch <= '~') || ch == '\t' || ch == '\n');
}


int getIpInt (char * msg, int & flag)
{
	string str;
	WINDOW *mesgWin = newwin (6, 35, LINES / 2 - 5, COLS / 2 - 15);
	PANEL *mesgPan = new_panel(mesgWin);
	wbkgd (mesgWin, COLOR_PAIR(diacolr));
	box (mesgWin, ACS_VLINE, ACS_HLINE);
	int len = strlen (msg);
	mvwprintw(mesgWin, 2, 3, "%s", msg);
	WINDOW *ipWin = derwin (mesgWin, 1, 15, 2, 4 + len);
	WINDOW *ok = derwin (mesgWin, 1, 10, 4, 3);
	WINDOW *cancel = derwin (mesgWin, 1, 10, 4, 22);
	keypad(ipWin, 1);
	wbkgd(ipWin, COLOR_PAIR(inpcolr));
	wbkgd (ok, COLOR_PAIR(butcolr));
	wbkgd (cancel, COLOR_PAIR(butcolr));	
	wattron (ok, A_BOLD);
	mvwprintw(ok, 0, 0, ">   OK   <");
	mvwprintw(cancel, 0, 0, "  CANCEL  ");	
	top_panel(mesgPan);
	update_panels();
	doupdate();
	wmove(ipWin, 0, 0);
	wrefresh (ipWin);
	wrefresh(ok);
	wrefresh(cancel);
	int retVal = OK;
	int ch = mvwgetch (ipWin, 0, 0);
	int loc = 0;
	while (ch != '\n' && ch != 27) {
		if (loc == 0)
			curs_set (1);
		else curs_set (0);	
		if (isdisplay (ch) && !loc && ch != '\n' && ch != '\t') {
			waddch (ipWin, ch);
			str.push_back (ch);
		}
		else if (ch == KEY_BACKSPACE && !loc)
			if (str.length()) {
				str.erase(str.begin() + str.length() - 1);
				mvwprintw(ipWin, 0, 0, "%-*s ", str.length() + 1, (char*) str.c_str());
			}
		switch (ch) {
				case KEY_RIGHT:
					wattroff (ok, A_BOLD);
					mvwprintw(ok, 0, 0, "    OK    ");
					wrefresh(ok);
					wattron (cancel, A_BOLD);
					mvwprintw(cancel, 0, 0, "> CANCEL <");	
					wrefresh(cancel);
					retVal = CANCEL;
					break;
				case KEY_LEFT:
					wattron (ok, A_BOLD);
					mvwprintw(ok, 0, 0, ">   OK   <");
					wattroff(cancel, A_BOLD);
					mvwprintw(cancel, 0, 0, "  CANCEL  ");	
					wrefresh(cancel);
					wrefresh(ok);
					retVal = OK;
					break;
			}
		ch = mvwgetch(ipWin, 0, str.length());
		if (ch == 27) {
			retVal = CANCEL;
			break;
		}
	}
	del_panel(mesgPan);
	delwin (ipWin);
	delwin(ok);
	delwin(cancel);
	delwin(mesgWin);
	update_panels();
	doupdate();
	noecho();
	flag = retVal;
	int num = 0; 
	int numDig = str.length();
	for (int i = 0; i <= numDig - 1; i++) 
		if (str[i] >= '0' && str[i] <= '9')
			num = num * 10 + (str[i] - '0');
		else return 0;
	return num; 
}
